
! Copyright (C) 2002 - Observatoire de Paris/Meudon
!                      Universite Paris 7
!                      CNRS (France)
!                      Contact : Jacques.Lebourlot@obspm.fr
!
! This program is free software; you can redistribute it and/or
! modify it under the terms of the GNU General Public License version 2
! as published by the Free Software Foundation.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


MODULE RF_TOOLS

!*****************************************************************************************
! Radiation Field manipulations tools
!*****************************************************************************************

  USE constantes

  IMPLICIT NONE

! RF_TYPE is used for all wavelength dependant quantities.
!         indexes go from lower to upper (the user does have to care if lower = 0 or not)
!         Values are in member Val. They must be ordered in increasing order of WaveLength
!         Ord is used to retrieve the Wave Length value from object rf_wl
!         rf_wl%Ord(i) is the rank at which rf_wl%val(i) has been entered in rf_wl
!         A specific vector (wl_pointer) is used to retrieve i from a given iord
!         Thus rf_wl may been unevenly spaced, and values entered in ony order

! A special set of subroutines and functions is provided to manipulate RF_TYPE objects
!
!   SUBROUTINE rf_create (rf_xx, ll, uu)
!              creates object rf_xx, and allocates Val and Ord from ll to uu
!   SUBROUTINE rf_free (rf_xx)
!              deallocates rf_xx%Val and rf_xx%Ord
!   SUBROUTINE rf_extand (rf_xx, x, add_1, iloc, iord)
!              add values x in rf_xx%Val. Val must be orderered in ascending order
!              x is inserted at its proper rank and the routine returns its location (iloc)
!              and its ordering number (iord). Note that iloc may vary as new values are added
!              but iord never changes.
!              If x is allready in rf_xx, add_1 returns with value 0. If not, add_1 = 1
!              rf_extand was created and is currently used only to add new wavelengths in
!              rf_wl
!   SUBROUTINE rf_locate (rf_xx, x, j)
!              returns index j such that rf_xx%Val(i) <= x < rf_xx%Val(i+1)
!              if x is outside rf_xx%Val, returns either lower or upper as appropriate
!              rf_xx%Val must be ordered in ascending order.
!   SUBROUTINE Shell_sort (rf_xxin, rf_xxout)
!              Sorts rf_object rf_xxin in ascending order
!   SUBROUTINE rf_revers (rindex)
!              Builds vector wl_pointer used to retrieve location of a given wavelength
!              from its ordering number.
!   REAL(KIND=DP) FUNCTION  rf_integration (xx, ffx)
!              Integrates ffx over xx using a trapezium rule.
!              xx and ffx are both rf_objects and must have the same size
!   REAL(KIND=DP) FUNCTION  sect_int (sect, rf_fn, rindex, wl_thresh)
!              Integrates the product sect * rf_fn over wavelength up to wl_thresh.
!              sect may have fewer points than rf_wl, but all wl must be in rf_wl (usually
!              added via rf_extand). rf_fn must have the same number of points as rf_wl
!              (usually, Radiation field intensity if sect is a cross section).
!              sect is interpolated linearly between two consecutive points, and the
!              integral between those points is performed by rf_integration.
!

  TYPE RF_TYPE
      INTEGER                                          :: lower             !  Lower index
      INTEGER                                          :: upper             !  Upper index
      REAL (KIND=DP), DIMENSION (:), POINTER           :: Val               !  Values....
!     REAL (KIND=DP), DIMENSION (:), POINTER           :: Val => NULL()     !  Values....
      INTEGER, DIMENSION (:), POINTER                  :: Ord               !  Order
  END TYPE RF_TYPE

! Wavelength and Radiation field

    TYPE (RF_TYPE), PUBLIC                             :: rf_wl                 !  Radiation field grid (WaveLength)
    TYPE (RF_TYPE), PUBLIC                             :: rf_dray               !  Radiation field intensity(erg/cm^3/A)
    TYPE (RF_TYPE), PUBLIC                             :: rf_draypr             !  Radiation field intensity (previous)
    INTEGER, DIMENSION (:), POINTER, PUBLIC            :: wl_pointer            !  Reverse ordering of wavelength

    REAL (KIND=DP), PUBLIC                             :: wlth0                 !  Start of Wavelength grid (H ionisation)
    REAL (KIND=DP), PUBLIC                             :: wlthm                 !  End of Wavelength grid
    REAL (KIND=DP), PUBLIC                             :: wl_hab = 2400.0_DP    !  Limit of Habing Field
    REAL (KIND=DP), PUBLIC                             :: wl_6ev = 2066.37_DP   !  Limit of Habing Field
    REAL (KIND=DP), PUBLIC                             :: pwlg = 16.0_DP        !  Initial WL step
    INTEGER, PUBLIC                                    :: nwlg = 200            !  Initial number of WL points
    INTEGER, PUBLIC                                    :: mfgk                  !  Number of lines (H2 + CO + ...) in FGK transfer
    REAL (KIND=DP), PUBLIC, DIMENSION (:), ALLOCATABLE :: fgkwl                 !  Wavelength of "FGK lines"
    REAL (KIND=DP), PUBLIC, DIMENSION (:), ALLOCATABLE :: fgkrf                 !  Line absorption (direct)
    REAL (KIND=DP), PUBLIC, DIMENSION (:), ALLOCATABLE :: fgkrfr                !  Line absorption (reverse)
    REAL (KIND=DP), PUBLIC, DIMENSION (:), ALLOCATABLE :: fgkgrd                !  Line self-shielding

! Photo-ionisation and photo-dissociation related quantities

    REAL (KIND=DP), PUBLIC, DIMENSION (0:noptm)        :: pdish2                !  H2 Photdissociation rate at iopt
    REAL (KIND=DP), PUBLIC, DIMENSION (0:noptm)        :: pdishd                !  HD Photdissociation rate at iopt

    REAL (KIND=DP), PUBLIC, DIMENSION (0:noptm)        :: pdico0                !  CO Photdissociation rate at iopt
    REAL (KIND=DP), PUBLIC, DIMENSION (0:noptm)        :: pdico1                !  13CO Photdissociation rate at iopt
    REAL (KIND=DP), PUBLIC, DIMENSION (0:noptm)        :: pdico2                !  C18O Photdissociation rate at iopt
    REAL (KIND=DP), PUBLIC, DIMENSION (0:noptm)        :: pdico3                !  13C18O Photdissociation rate at iopt

    TYPE (RF_TYPE), PUBLIC                             :: rf_sig_C              !  Carbon Photo-ionisation
    REAL (KIND=DP)                                     :: wl_cut_C = 1101.0_DP  !  C ionisation cut off
    REAL (KIND=DP)                                     :: sig_C = 1.6e-17_DP    !  C ionisation cross-section (Cte)
    REAL (KIND=DP), PUBLIC, DIMENSION (0:noptm)        :: pdicat                !  C Photionisation rate at iopt

    REAL (KIND=DP)                                     :: wl_cut_S              !  C ionisation cut off
    TYPE (RF_TYPE), PUBLIC                             :: rf_sig_S              !  Sulphur Photo-ionisation (TopBase)
    REAL (KIND=DP), PUBLIC, DIMENSION (0:noptm)        :: pdisat                !  S Photionisation rate at iopt

    TYPE (RF_TYPE), PUBLIC                             :: rf_exter              !  External radiation field (if any)
    INTEGER                                            :: nbpt, nskip

    REAL (KIND=DP), PUBLIC                             :: int_rad0              !
    REAL (KIND=DP), PUBLIC                             :: int_rad1              !
    REAL (KIND=DP), PUBLIC                             :: int_rad2              !

! Grain photo-electric efficiency and extinction curve

    TYPE(RF_TYPE), PUBLIC                              :: rf_abspou             !  Absorption by dust (Fitzpatrick)

    REAL (KIND=DP), PUBLIC, DIMENSION(:,:), ALLOCATABLE :: si_QU, si_QUl, si_QUl2
    REAL (KIND=DP), PUBLIC, DIMENSION(:), ALLOCATABLE  :: si_
    TYPE(RF_TYPE),PUBLIC, DIMENSION(:), ALLOCATABLE    :: sigabs
    TYPE(RF_TYPE),PUBLIC                               :: dust_wl

! This is used for photo-electric heating (Bakes & Tielens formalism) in stv5c.f)

  TYPE(RF_TYPE), public                                :: bt_int                ! Integrand (without sigma)

! This table saves the radiation field at some "imposed" Av for output

  REAL (kind=dp), public, DIMENSION(:,:), ALLOCATABLE  :: rfield

CONTAINS

! %%%%%%%%%%%%%%%%%%
  SUBROUTINE rf_init
! %%%%%%%%%%%%%%%%%%
! purpose : Initialise model dependant quantities related to Radiation Field (RF)
!   * Wavelength grid                  : rf_wl (need not have constant step)
!   * RF intensity                     : rf_dray (currently Draine field only)
!   * Dust absorption curve            : rf_abspou (Fitzpatrick fit)

!---------------------------Declaration of parameters-------------------------
    USE constantes
    USE auxiliaries
    USE inoutput
    USE PROFIL
    USE DRAINE_DUST

    IMPLICIT NONE

    REAL (KIND=DP)                             :: wl_cut_D = 2000.0_DP !  Draine field cutoff wavelength
    INTEGER                                    :: I_cut_D              !  Draine field cutoff position in grid
    REAL (KIND=DP)                             :: wl_cut_F             !  Fitzpatrick cutoff wavelength
    INTEGER                                    :: I_cut_F              !  Fitzpatrick cutoff position in grid
    INTEGER                                    :: mm, nnp
    INTEGER                                    :: i, j, nb_pt
    INTEGER                                    :: ll, uu
    INTEGER                                    :: add_1, iloc, iord    !  Used by rf_extand
    REAL (KIND=DP)                             ::  Energy, sigm
    REAL (KIND=DP)                             :: wl, wlo, rf, rfo
    TYPE (RF_TYPE)                             :: rf_tmp
    REAL (KIND=DP)                             :: summ_f, summ_f2
    INTEGER                                    :: npsec
    parameter (npsec = 100)

    INTEGER                                    :: nf, nfb, m, ilamb
    REAL(KIND=DP)                              :: sigbg, sigvsg, sigpah, wlnf, wl1
    REAL(KIND=DP)                              :: wl0, dopl, dwl
    REAL(KIND=DP), PARAMETER                   :: wlrinc = 1.2_dp
    INTEGER                                    :: njl, nvl
    INTEGER                                    :: jwlmin, jwlmax

!-------------- Execution  ------------------------------------------------------------

! Compute H ionisation potential

    hiop = 2.0_DP * xpi*xpi * xqe**4 * xme / (xhp*xhp)
    wlth0 = 1.0e8 * xhp * clum / hiop
    wlthm = wlth0 + nwlg * pwlg

    print *, "  wlth0 =", wlth0, " wlthm =", wlthm

! Initialisation of wavelenghts
! rf_wl%Ord is a permutation of (rf_wl%lower : rf_wl%upper) giving the
!           order in which rf_wl%Val have been entered

    CALL rf_create (rf_wl, 0, nwlg)
    rf_wl%Val = wlth0 + pwlg * (/ (I, I=0,nwlg) /)
    rf_wl%Ord = (/ (I, I=0,nwlg) /)

!  Add all "special values"

!    print *
!    print *,'nwlg before :', nwlg

!  Habing field limit

    CALL rf_extand (rf_wl, wl_hab, add_1, iloc, iord)
    nwlg = nwlg + add_1

!  6 eV field limit

    CALL rf_extand (rf_wl, wl_6ev, add_1, iloc, iord)
    nwlg = nwlg + add_1

!  Fitzpatrick & Massa extinction curve

    wl_cut_F = 1.0e4_DP / 5.9_DP
    CALL rf_extand (rf_wl, wl_cut_F, add_1, iloc, iord)
    nwlg = nwlg + add_1
!    print *, iloc, iord, wl_cut_F

! Draine Field cut-off

     CALL rf_extand (rf_wl, wl_cut_D, add_1, iloc, iord)
     nwlg = nwlg + add_1
!     print *, iloc, iord, wl_cut_D
!     print *, 'nwlg after :', nwlg

! -----------------------------------------------------------------------------

! Read the cross sections as given by the TopBase data base.

! * - Carbon

     CALL rf_create (rf_sig_C,1,2)

     CALL rf_extand (rf_wl, wlth0, add_1, iloc, iord)
     nwlg = nwlg + add_1
     rf_sig_C%Val(1) = sig_C
     rf_sig_C%Ord(1) = iord

     CALL rf_extand (rf_wl, wl_cut_C, add_1, iloc, iord)
     nwlg = nwlg + add_1
     rf_sig_C%Val(2) = sig_C
     rf_sig_C%Ord(2) = iord

! * - Sulphur

     fichier = TRIM(data_dir)//'sig_S.dat'
     OPEN(iread2, file = fichier, status = 'old')
! First count number of point above threshold

     READ(iread2,"(7/)")
     READ(iread2,*) nb_pt
     DO i =1,nb_pt
       READ (iread2,*) Energy
       wl = ryd2ang / Energy                      ! Energy or wavelength in Angstrom
       if (wl < wlth0) then
         exit
       endif
     END DO
     nb_pt = i
     close (iread2)
! Then read correct number of cross section and interpolate last to wlth0

     CALL rf_create (rf_sig_S,1,nb_pt)
     OPEN(iread2, file = fichier, status = 'old')
     READ(iread2,"(8/)")

     DO i = 1, nb_pt-1
       READ (iread2,*) Energy, sigm
       wl = ryd2ang / Energy
       if (i == 1) then
         wl_cut_S = wl
       endif
       CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
       nwlg = nwlg + add_1
       rf_sig_S%Val(nb_pt-i+1) = 1.e-18_DP * sigm   ! sigm's unit is cm^-2
       rf_sig_S%Ord(nb_pt-i+1) = iord
     END DO

     READ (iread2,*) Energy, sigm
     sigm = sigm * 1.e-18_DP
     wl = ryd2ang / Energy
     CALL rf_locate (rf_wl, wlth0, i)
     rf_sig_S%Val(1) = sigm + (rf_sig_S%Val(2) - sigm) * &
                      (wlth0 - wl) / (rf_wl%Val(rf_wl%upper) - wl)
     rf_sig_S%Ord(1) = rf_wl%Ord(i)

     close (iread2)

! 17 III 2004 : read external radiation field

     if (srcpp(1:2) == 'F_') then
       print *, srcpp
       fichier = TRIM(data_dir)//srcpp
       open(iread2, file = fichier, status = 'old')
       read (iread2,*) rstar
       read (iread2,*) teff
       print *, rstar, teff
       rstar = rstar * rsun
       read (iread2,*) nb_pt
       read (iread2,*)
       print *, nb_pt

       nbpt = 0
       nskip = 0
       do i =1,nb_pt
         read (iread2,*) wl
         wl = wl * 1.0e1_dp      ! convert to Angstrom
         if (wl < wlth0) then
           nskip = nskip + 1
           cycle
         endif
         nbpt = nbpt + 1
         if (wl > wlthm) then
           exit
         endif
       end do
       close (iread2)
       nbpt = nbpt + 1
       print *, nbpt

       call rf_create (rf_exter,1,nbpt)
       open(iread2, file = fichier, status = 'old')
       read(iread2,"(3/)")
       wl = 0.0_dp
       rf = 0.0_dp
       do i=1,nskip+1
         wlo = wl
         rfo = rf
         read (iread2,*) wl, rf
         wl = wl * 1.0e1_dp      ! convert to Angstrom
!        rf = rf * 0.1 / clum    ! convert to erg cm-3 Angstr-1
         rf = rf * 0.1 * 4.0_dp * xpi / clum    ! convert to erg cm-3 Angstr-1
       end do

       CALL rf_locate (rf_wl, wlth0, i)
       rf_exter%Val(1) = rfo + (rf - rfo) * &
                        (wlth0 - wlo) / (wl - wlo)
       rf_exter%Ord(1) = rf_wl%Ord(i)
       CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
       nwlg = nwlg + add_1
       rf_exter%Val(2) = rf
       rf_exter%Ord(2) = iord

       do i =3,nbpt
         wlo = wl
         rfo = rf
         read (iread2,*) wl, rf
         wl = wl * 1.0e1_dp      ! convert to Angstrom
         rf = rf * 0.1 * 4.0_dp * xpi / clum    ! convert to erg cm-3 Angstr-1
         if (wl > wlthm) then
           exit
         endif
         CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
         nwlg = nwlg + add_1
         rf_exter%Val(i) = rf
         rf_exter%Ord(i) = iord
       end do
       close (iread2)

       CALL rf_locate (rf_wl, wlthm, i)
       rf_exter%Val(nbpt) = rfo + (rf - rfo) * &
                        (wlthm - wlo) / (wl - wlo)
       rf_exter%Ord(nbpt) = rf_wl%Ord(i)

     endif

! 24 VII 2003
!       Grains absorption cross sections are interpolated from Draines data
!       for each radius.

     CALL READ_DRAINE_FILES

     CALL MIXGRSI (wlth0, wlthm, jwlmin, jwlmax)

     allocate (sigabs(npg))
     do mm=1,npg
       NULLIFY (sigabs(mm)%val,sigabs(mm)%ord)                     !SG
       CALL rf_create(sigabs(mm),1,jwlmax-jwlmin+2)
     end do

     DO i=1,jwlmax-jwlmin+2
       if (i == 1) then
         wl = wlth0
       else if (i == jwlmax-jwlmin+2) then
         wl = wlthm
       else
         wl = wavgrd(i+jwlmin-1)
       endif
       CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
       nwlg = nwlg + add_1
       do mm=1,npg
         sigabs(mm)%Val(i) = Qabs(i,mm)
         sigabs(mm)%Ord(i) = iord
       end do
     END DO

! 2 -  build reverse indexing of wavelength

     CALL rf_revers (wl_pointer)

! The wavelength grid need now be extended to give accurate integration
! 3 -  integrate test function over wavelength

     summ_f = sect_int (rf_sig_S, rf_wl, wl_pointer, wl_cut_S)

!     print *, "  summ_f :", summ_f

! 4 -  iteratively refine wavelength grid until continuum integration is "good"

     summ_f2 = 0.5_DP * summ_f
     j = 0
      print *, "   nwlg 1:", nwlg
     do while (abs((summ_f-summ_f2)/summ_f) > 1.0e-5)
       j = j + 1

       pwlg = pwlg * 0.5_DP
       i = 0
       do while (i < nwlg-1)
         if ((rf_wl%Val(i+1)-rf_wl%Val(i)) > pwlg) then
           wl = 0.5_DP * (rf_wl%Val(i+1)+rf_wl%Val(i))
           CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
           nwlg = nwlg + add_1
         else
           i = i + 1
         endif
       end do
        print *, "   nwlg",j,":", nwlg

       CALL rf_revers (wl_pointer)
       summ_f2 = summ_f
       summ_f = sect_int (rf_sig_S, rf_wl, wl_pointer, wl_cut_S)
!       print *, "  summ_f :", summ_f2, summ_f

     end do

! 21 V 2003 - Add line wavelenght in wavelength grid
! 1 - Atomic hydrogen (Lyman lines)
!     parameter "dopl" is the step in lambda
!                      mesured in Doppler width

     if (jlyman == 1) then
     do i=1,nraih
       wl0 = praih(i,2)
       dopl = 0.3_dp * wl0 * vturb / clum
       CALL rf_extand (rf_wl, wl0, add_1, iloc, iord)
       nwlg = nwlg + add_1

       dwl = 0.2_dp * dopl
       wl = wl0
       do
       if (abs(wl-wl0)/dopl < wlrwid) then
         wl1 = rf_wl%Val(iloc+1)
         if ((wl1-wl) > 2.0_dp*dwl) then
           wl = wl + dwl
           CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
           nwlg = nwlg + add_1
         else if ((wl1-wl) > dwl) then
           wl = wl + (wl1-wl) * 0.5_dp
           CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
           nwlg = nwlg + add_1
           wl = wl1
           iloc = iloc + 1
         else
           wl = wl1
           iloc = iloc + 1
         endif
         dwl = wlrinc * dwl
       else
         exit
       endif
       end do

       call rf_locate (rf_wl, wl0, iloc)

       dwl = 0.2_dp * dopl
       wl = wl0
       do
       if (abs(wl-wl0)/dopl < wlrwid .AND. &
           wl > wlth0) then
         wl1 = rf_wl%Val(iloc-1)
         if ((wl-wl1) > 2.0_dp*dwl) then
           wl = wl - dwl
           CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
           nwlg = nwlg + add_1
         else if ((wl-wl1) > dwl) then
           wl = wl - (wl-wl1) * 0.5_dp
           CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
           nwlg = nwlg + add_1
           wl = wl1
           iloc = iloc - 1
         else
           wl = wl1
           iloc = iloc - 1
         endif
         dwl = wlrinc * dwl
       else
         exit
       endif
       end do

     end do
     endif

!  Add lines of H2, only Jlow = 0 or 1

     do i=1,nrh2bt
       nvl = irah2b(i,1)
       njl = irah2b(i,2)
       if (nvl > 0 .OR. njl >= jfgkh2) then
         cycle
       endif
       wl0 = prah2b(i,2)
       dopl = 0.3_dp * wl0 * vturb / clum
       CALL rf_extand (rf_wl, wl0, add_1, iloc, iord)
       nwlg = nwlg + add_1

       dwl = 0.2_dp * dopl
       wl = wl0
       do
       if (abs(wl-wl0)/dopl < wlrwid) then
         wl1 = rf_wl%Val(iloc+1)
         if ((wl1-wl) > 2.0_dp*dwl) then
           wl = wl + dwl
           CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
           nwlg = nwlg + add_1
         else if ((wl1-wl) > dwl) then
           wl = wl + (wl1-wl) * 0.5_dp
           CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
           nwlg = nwlg + add_1
           wl = wl1
           iloc = iloc + 1
         else
           wl = wl1
           iloc = iloc + 1
         endif
         dwl = wlrinc * dwl
       else
         exit
       endif
       end do

       call rf_locate (rf_wl, wl0, iloc)

       dwl = 0.2_dp * dopl
       wl = wl0
       do
       if (abs(wl-wl0)/dopl < wlrwid .AND. &
           wl > wlth0) then
         wl1 = rf_wl%Val(iloc-1)
         if ((wl-wl1) > 2.0_dp*dwl) then
           wl = wl - dwl
           CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
           nwlg = nwlg + add_1
         else if ((wl-wl1) > dwl) then
           wl = wl - (wl-wl1) * 0.5_dp
           CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
           nwlg = nwlg + add_1
           wl = wl1
           iloc = iloc - 1
         else
           wl = wl1
           iloc = iloc - 1
         endif
         dwl = wlrinc * dwl
       else
         exit
       endif
       end do

     end do

     do i=1,nrh2ct
       nvl = irah2c(i,1)
       njl = irah2c(i,2)
       if (nvl > 0 .OR. njl >= jfgkh2) then
         cycle
       endif
       wl0 = prah2c(i,2)
       dopl = 0.3_dp * wl0 * vturb / clum
       CALL rf_extand (rf_wl, wl0, add_1, iloc, iord)
       nwlg = nwlg + add_1

       dwl = 0.2_dp * dopl
       wl = wl0
       do
       if (abs(wl-wl0)/dopl < wlrwid) then
         wl1 = rf_wl%Val(iloc+1)
         if ((wl1-wl) > 2.0_dp*dwl) then
           wl = wl + dwl
           CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
           nwlg = nwlg + add_1
         else if ((wl1-wl) > dwl) then
           wl = wl + (wl1-wl) * 0.5_dp
           CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
           nwlg = nwlg + add_1
           wl = wl1
           iloc = iloc + 1
         else
           wl = wl1
           iloc = iloc + 1
         endif
         dwl = wlrinc * dwl
       else
         exit
       endif
       end do

       call rf_locate (rf_wl, wl0, iloc)

       dwl = 0.2_dp * dopl
       wl = wl0
       do
       if (abs(wl-wl0)/dopl < wlrwid .AND. &
           wl > wlth0) then
         wl1 = rf_wl%Val(iloc-1)
         if ((wl-wl1) > 2.0_dp*dwl) then
           wl = wl - dwl
           CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
           nwlg = nwlg + add_1
         else if ((wl-wl1) > dwl) then
           wl = wl - (wl-wl1) * 0.5_dp
           CALL rf_extand (rf_wl, wl, add_1, iloc, iord)
           nwlg = nwlg + add_1
           wl = wl1
           iloc = iloc - 1
         else
           wl = wl1
           iloc = iloc - 1
         endif
         dwl = wlrinc * dwl
       else
         exit
       endif
       end do

     end do

     print *, "   nwlg =", nwlg
!    print *, "   pwlg =", pwlg

     if (rf_wl%upper /= nwlg) then
       print *, "Probleme"
     endif

     CALL rf_revers (wl_pointer)

!---------------------------------------------------------------------------------------------------------

! -----------------------------------------------------------------------------
!    From here, the set of wavelength is complete and should not be modified !
! -----------------------------------------------------------------------------

!----------------------------------------------------------------------------------------------------------
! Initialisation of the radiation Field.
! using Draine Formula as revised by M. Kopp

    CALL rf_create (rf_dray, 0, nwlg)
    CALL rf_create (rf_draypr, 0, nwlg)

! Draine Formula

    CALL rf_locate (rf_wl, wl_cut_D, I_cut_D)

    rf_dray%Val(0:I_cut_D) = 6.36E7_DP / rf_wl%Val(0:I_cut_D)**4 &
                         - 1.0237E11_DP / rf_wl%Val(0:I_cut_D)**5 &
                         + 4.0812E13_DP / rf_wl%Val(0:I_cut_D)**6

    rf_dray%Val(I_cut_D+1:nwlg) = 1.38243E-5_DP &
                                * rf_wl%Val(I_cut_D+1:nwlg)**(-0.3_DP)

    rf_dray%Val = rf_dray%Val / clum
    rf_dray%Ord = (/ (I, I=rf_dray%lower,rf_dray%upper) /)

    rf_draypr%Val = rf_dray%Val
    rf_draypr%Ord = rf_dray%Ord

!---------------------------------------------------------------------------------------------------------

! Fitzpatrick absorption curves

!  courbe d extinction d apres Fitzpatrick

     CALL rf_create (rf_abspou, 0, nwlg)
     CALL rf_create (rf_tmp, 0, nwlg)
     rf_tmp%Val = 1.0e4_DP / rf_wl%Val

     CALL rf_locate (rf_wl, wl_cut_F, I_cut_F)


     rf_abspou%Val =  cc1 + cc2 * rf_tmp%Val + cc3 * (rf_tmp%Val * rf_tmp%Val / &
        ((rf_tmp%Val * rf_tmp%Val - y0 * y0)**2 + xgam * xgam * rf_tmp%Val * rf_tmp%Val))

     rf_abspou%Val(0:I_cut_F) = rf_abspou%Val(0:I_cut_F) &
        +  cc4 * (0.539_DP * (rf_tmp%Val(0:I_cut_F)-5.9_DP)**2 &
        + 0.0564_DP * (rf_tmp%Val(0:I_cut_F)-5.9_DP)**3)

     rf_abspou%Val = 1.0_DP + rf_abspou%Val / rrr
     rf_abspou%Ord = (/ (I, I=rf_abspou%lower,rf_abspou%upper) /)

     CALL rf_free (rf_tmp)

! Compute photoionisation at edge

     CALL rf_create (rf_tmp, 0, nwlg)
     rf_tmp%Val = rf_wl%Val * rf_dray%Val
     pdicat(0) = sect_int (rf_sig_C, rf_tmp, wl_pointer, wl_cut_C) / (1.0e8_DP * xhp)
     pdisat(0) = sect_int (rf_sig_S, rf_tmp, wl_pointer, wl_cut_S) / (1.0e8_DP * xhp)

!    print *, " C:", pdicat(0), ",  S:", pdisat(0)

! Create RF_TYPE objects and tables used in fbth (stv5c.f)

     ALLOCATE (si_(rf_wl%lower:rf_wl%upper))
     ALLOCATE (si_QU(npg,rf_wl%lower:rf_wl%upper))
     ALLOCATE (si_QUl(npg,rf_wl%lower:rf_wl%upper))
     ALLOCATE (si_QUl2(npg,rf_wl%lower:rf_wl%upper))

     do mm=1,npg
       rf_tmp%Ord = rf_dray%Ord
       rf_tmp%Val = rf_dray%Val
       call sect_pri(sigabs(mm), rf_tmp, wl_pointer, si_, nwlg)
       si_QU(mm,:) = si_
       rf_tmp%Val = rf_tmp%Val * rf_wl%Val
       call sect_pri(sigabs(mm), rf_tmp, wl_pointer, si_, nwlg)
       si_QUl(mm,:) = si_
       rf_tmp%Val = rf_tmp%Val * rf_wl%Val
       call sect_pri(sigabs(mm), rf_tmp, wl_pointer, si_, nwlg)
       si_QUl2(mm,:) = si_
     end do

     wl = 0.5_dp * (wlth0+wlthm)
     dwl = sect_int(sigabs(1), rf_dray, wl_pointer, wl)
     call rf_locate (rf_wl, wl, mm)
     print *, dwl, si_QU(1,mm)

     call rf_create(bt_int,rf_wl%lower,rf_wl%upper)

! Create Radiation field table for output only

     ALLOCATE (rfield(0:n_imp,rf_wl%lower:rf_wl%upper))

     CALL rf_free (rf_tmp)
 
  END SUBROUTINE rf_init

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  SUBROUTINE sect_pri (sect, rf_fn, rindex, sect_p, nwlg)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! Integrate The product sect * rf_fn over lambda
!           and save each partial sum from lamb1 to lamb_i

  IMPLICIT NONE

  INTEGER                             :: nwlg
  TYPE(RF_TYPE), intent(in)           :: sect       ! Section to integrate
  TYPE(RF_TYPE), intent(in)           :: rf_fn      ! Weight function
  INTEGER, POINTER, DIMENSION(:)      :: rindex     ! Reverse ordering index
  REAL (KIND = DP), DIMENSION(0:nwlg) :: sect_p     ! partial integrated section

  INTEGER                        :: i, j
  INTEGER                        :: ll, uu
  REAL (KIND = DP)               :: s1, s2, l1, l2
  REAL (KIND = DP)               :: s, l, so, lo

  sect_p = 0.0_dp
  sect_p(rindex(sect%Ord(sect%lower))) = 1.0e-40_dp

  do i=sect%lower, sect%upper-1
    ll = rindex(sect%Ord(i))
    l1 = rf_wl%Val(ll)
    s1 = sect%Val(i)

    uu = rindex(sect%Ord(i+1))
    l2 = rf_wl%Val(uu)
    s2 = sect%Val(i+1)

    lo = l1
    so = s1
    do j=ll,uu-1
      l = rf_wl%Val(j+1)
      s = s1 + (s2 - s1) * (l - l1) / (l2 - l1)
      sect_p(j+1) = sect_p(j) + 0.5_dp * (l - lo) &
                  * (so * rf_fn%Val(j) + s * rf_fn%Val(j+1))
      lo = l
      so = s
    end do

  end do

  END SUBROUTINE sect_pri

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 REAL(KIND=DP) FUNCTION  sect_int (sect, rf_fn, rindex, wl_thresh)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!---------------------------------------------------------------------------------------------
! Integrate The product sect * rf_fn over lambda
!           From lamb1 to lamb2
! sect is interpolated between two successive points
! rf_fn must be known at every point in lambda
!---------------------------------------------------------------------------------------------

  IMPLICIT NONE

  TYPE(RF_TYPE), intent(in)      :: sect       ! Section to integrate
  TYPE(RF_TYPE), intent(in)      :: rf_fn      ! Weight function
  INTEGER, POINTER, DIMENSION(:) :: rindex     ! Reverse ordering index
  REAL (KIND = DP)               :: wl_thresh  ! Threshold

  INTEGER                        :: i, ll, uu
  INTEGER                        :: I_thresh
  REAL (KIND = DP)               :: sumt, sump
  REAL (KIND = DP)               :: s1, s2, l1, l2
  REAL (KIND = DP)               :: i_int, j_int
  TYPE (RF_TYPE)                 :: rf_xx
  TYPE (RF_TYPE)                 :: rf_fxx

  call rf_locate (rf_wl, wl_thresh, I_thresh)

  sumt = 0.0_DP
  do i=sect%lower, sect%upper-1

    ll = rindex(sect%Ord(i))
    l1 = rf_wl%Val(ll)
    s1 = sect%Val(i)

    uu = rindex(sect%Ord(i+1))
    l2 = rf_wl%Val(uu)
    s2 = sect%Val(i+1)

    if (uu > I_thresh) then
      uu = I_thresh
      s2 = s1 + (s2 - s1) * (rf_wl%Val(uu) - l1) / (l2 - l1)
      l2 = rf_wl%Val(uu)
    endif

    if (l2 > l1) then

      CALL rf_create (rf_xx, ll, uu)
      CALL rf_create (rf_fxx, ll, uu)

      rf_xx%Val = rf_wl%Val(ll:uu)
      rf_fxx%Val = rf_fn%Val(ll:uu)
      i_int = rf_integration (rf_xx, rf_fxx)
      rf_fxx%Val = rf_fxx%Val * rf_xx%Val
      j_int = rf_integration (rf_xx, rf_fxx)

      sump = (j_int * (s1-s2) - i_int * (s1*l2-s2*l1)) / (l1-l2)    ! Analytical integration of sect
      sumt = sumt + sump
      CALL rf_free (rf_fxx)
      CALL rf_free (rf_xx)
    else
      exit
    endif

  end do

  sect_int = sumt

END FUNCTION sect_int

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 REAL(KIND=DP) FUNCTION  rf_integration (xx, ffx)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!---------------------------------------------------------------------------------------------
! Integrate Function using the trapezium rules.
! With uneven step
!---------------------------------------------------------------------------------------------

  IMPLICIT NONE

  TYPE(RF_TYPE), intent(in) :: xx               ! Abscissa
  TYPE(RF_TYPE), intent(in) :: ffx              ! Y-axis
  INTEGER                   :: l, u, i
  l = xx%lower
  u = xx%upper

  rf_integration = 0.5_DP * SUM( (/ ((xx%Val(i+1) - xx%Val(i)) &
                                  * (ffx%Val(i+1) + ffx%Val(i)), i= l, u-1) /) )

END FUNCTION rf_integration

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     SUBROUTINE rf_create (rf_xx, ll, uu)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! Purpose: initialize upper and lower range of Val and Ord
!          and allocate the corresponding vector
! Note : Do NOT forget to deallocate temporary rf_objects !

  USE constantes

  IMPLICIT NONE

  TYPE(RF_TYPE), intent(inout) :: rf_xx
  INTEGER, intent (in)         :: ll, uu

  rf_xx%lower = ll
  rf_xx%upper = uu
!Choisir la ligne en fonction du compilateur
!pour pgi: allocated et pour absoft: associated
!Sous Windows (Visula Fortran) : associated
! IF ( .NOT. ALLOCATED(rf_xx%Val)) THEN
  IF ( .NOT. ASSOCIATED(rf_xx%Val)) THEN
    ALLOCATE (rf_xx%Val(ll:uu))
    ALLOCATE (rf_xx%Ord(ll:uu))
  ELSE
    PRINT *, "  Warning : Allocation is done by rf_create"
    PRINT *, "            Do not forget to deallocate if called from a loop"
    CALL rf_free (rf_xx)
    ALLOCATE (rf_xx%Val(ll:uu))
    ALLOCATE (rf_xx%Ord(ll:uu))
  ENDIF

  rf_xx%Val = 0.0_dp
  rf_xx%Ord = 0


  END SUBROUTINE rf_create

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     SUBROUTINE rf_free (rf_xx)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! Purpose: deallocate all allocatable bits of rf_xx

  IMPLICIT NONE
  TYPE(RF_TYPE), INTENT (INOUT) :: rf_xx

  DEALLOCATE (rf_xx%Val)
  DEALLOCATE (rf_xx%Ord)

  END SUBROUTINE rf_free

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     SUBROUTINE rf_revers (rindex)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! Purpose: Reverse ordering of wavelength

  IMPLICIT NONE
  INTEGER, POINTER, DIMENSION(:) :: rindex
  INTEGER :: i

     NULLIFY (rindex)
     ALLOCATE (rindex(rf_wl%lower:rf_wl%upper))

     do i=rf_wl%lower,rf_wl%upper
       rindex(rf_wl%Ord(i)) = i
     end do

  END SUBROUTINE rf_revers

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     SUBROUTINE rf_extand (rf_xx, x, add_1, iloc, iord)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
! Purpose: Add one value in rf_xx%Val if not already here
!          and update upper and rf_xx%Ord
! On output, iloc is the index of the added value in rf_xx
!            iord is the rank (order) of the added value

  IMPLICIT NONE

  TYPE(RF_TYPE), intent(inout) :: rf_xx
  REAL (KIND=DP), intent (in)  :: x
  INTEGER, INTENT (out)        :: add_1
  INTEGER, INTENT (out)        :: iloc
  INTEGER, INTENT (out)        :: iord

  TYPE(RF_TYPE)                :: rf_tmp
  INTEGER                      :: ll, uu
  INTEGER                      :: i

  ll = rf_xx%lower
  uu = rf_xx%upper

  CALL rf_locate (rf_xx, x, i)

! if x is already in rf_xx%Val, no need to extand

  IF (rf_xx%Val(i) == x) THEN
    add_1 = 0
    iloc = i
    iord = rf_xx%Ord(i)
    RETURN
  ENDIF

  CALL rf_create (rf_tmp, ll, uu)

  rf_tmp%lower = ll
  rf_tmp%upper = uu
  rf_tmp%Val = rf_xx%Val
  rf_tmp%Ord = rf_xx%Ord

  CALL rf_free (rf_xx)
  CALL rf_create (rf_xx, ll, uu+1)

  rf_xx%Val(ll:i) = rf_tmp%Val(ll:i)
  rf_xx%Ord(ll:i) = rf_tmp%Ord(ll:i)

  iloc = i + 1
  iord = uu + 1
  rf_xx%Val(iloc) = x
  rf_xx%Ord(iloc) = iord

  rf_xx%Val(i+2:uu+1) = rf_tmp%Val(i+1:uu)
  rf_xx%Ord(i+2:uu+1) = rf_tmp%Ord(i+1:uu)

  rf_xx%upper = uu + 1

  add_1 = 1
  CALL rf_free (rf_tmp)

  END SUBROUTINE rf_extand

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     SUBROUTINE Shell_sort (rf_xxin,rf_xxout)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!----------------------------------------------------------------------------------------------
! purpose: sort an rf_object in increasing ordre.
!      The sort methode is the "Shell" sort
!      Num Rec in Fortran, p 322  (second edition)
! On input rf_xxout%lower and rf_xxout%upper should have been
!      set to check compatibility of objects
! Outside Shell_sort, rf_xxin and rf_xxout can be the same object
!      In that case the sorted one replaces the unsorted one.
!----------------------------------------------------------------------------------------------

   IMPLICIT NONE

   TYPE(RF_TYPE), INTENT(IN)       :: rf_xxin
   TYPE(RF_TYPE), INTENT(INOUT)    :: rf_xxout

   TYPE(RF_TYPE)                   :: a
   INTEGER                         :: p
   REAL(KIND=DP)                   :: v
   INTEGER                         :: i, j, inc, n, l, u

!------------------------Execution------------------------------------------------------

   l = rf_xxin%lower
   u = rf_xxin%upper

  IF ( (rf_xxout%lower /= l) .OR. (rf_xxout%upper /= u) ) then
     print *, "  In Shell_sort: rf_xxin and rf_xxout should"
     print *, "                 have the same shape"
     print *, "  in :", l, u
     print *, "  out:", rf_xxout%lower, rf_xxout%upper
     stop
  ENDIF

  n = u - l + 1

  CALL rf_create (a, 1, n)
  a%Val(1:n) = rf_xxin%Val(l:u)
  a%Ord(1:n) = rf_xxin%Ord(l:u)

  inc = 1
  DO WHILE (inc <= n)
     inc = 3 * inc + 1
  END DO

  DO WHILE (inc > 1)
     inc = inc / 3

     DO i = inc+1, n
        v = a%Val(i)
        p = a%Ord(i)
        j = i
        DO WHILE ((a%Val(j-inc) > v))
           a%Val(j) = a%Val(j-inc)
           a%Ord(j) = a%Ord(j-inc)
           j = j - inc
           IF (j <= inc) EXIT
        END DO

        a%Val(j) = v
        a%Ord(j) = p
     END DO
  END DO

  rf_xxout%Val(l:u) = a%Val(1:n)
  rf_xxout%Ord(l:u) = a%Ord(1:n)

  CALL rf_free (a)

END SUBROUTINE Shell_sort

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  SUBROUTINE rf_locate (rf_xx, x, j)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
!  purpose : locate index j of rf_xx%Val such that
!            rf_xx%Val(j) <= x < rf_xx%Val(j+1)
!  need : rf_xx%Val should be sorted in ASCENDING order
!  Note : if x is outside rf_xx%Val, then return jl or ju

  IMPLICIT NONE

  TYPE (RF_TYPE), intent (in) :: rf_xx
  REAL (KIND=DP), intent (in) :: x
  INTEGER, intent (out)       :: j

  INTEGER                     :: jl, ju, jm

  jl = rf_xx%lower
  ju = rf_xx%upper

  IF (x >= rf_xx%Val(ju)) THEN
    j = ju
  ELSE IF (x < rf_xx%Val(jl)) THEN
    j = jl
  ELSE

    DO WHILE ((ju-jl) > 1)
      jm = (ju + jl) / 2
      IF (x < rf_xx%Val(jm)) THEN
         ju = jm
      ELSE
         jl = jm
      END IF
    END DO
    j = jl

  ENDIF

 END SUBROUTINE rf_locate

END MODULE RF_TOOLS
